#!/usr/bin/python
# -*- coding: utf-8 -*-

import sys
import random
import copy
import argparse

SUCESS_PROBABILITIES = True
PROB_DENOMINATOR = 100
MIN_PROB = 1
MAX_PROB = PROB_DENOMINATOR - 1
DEFAULT_EXPLOIT_COST = 1
MIN_COST = 1
MAX_COST = 1


class Vulnerability(object):

    def __init__(
        self,
        name,
        prob=PROB_DENOMINATOR,
        cost=DEFAULT_EXPLOIT_COST,
        ):

        self._name = name
        self._prob = prob
        self._cost = cost

    def get_name(self):
        return self._name

    def get_probability(self):
        return self._prob

    def get_cost(self):
        return self._cost

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other):
        return self._name == other._name

    def __repr__(self):
        return 'Vul(%s)' % self._name


class OperatingSystem(object):

    def __init__(self, name):
        self._name = name

    def get_name(self):
        return self._name

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other):
        return isinstance(other, OperatingSystem) and self._name \
            == other._name

    def __repr__(self):
        return 'OS(%s)' % self._name


class Application(object):

    def __init__(self, name, port):
        self._name = name
        self._port = port

    def get_name(self):
        return self._name

    def get_port(self):
        return self._port

    def __hash__(self):
        return hash(self._name)

    def __eq__(self, other):
        return self._name == other._name

    def __repr__(self):
        return 'App(%s)' % self._name


class ExploitDatabase(object):

    def __init__(self):
        self._numvulns = 0
        self._db = {}

    def add_os(self, os):
        if not os in self._db:
            self._db[os] = {}

    def add_application(self, os, prog):

        self.add_os(os)
        if not prog in self._db[os]:
            self._db[os][prog] = []
        return self._db[os][prog]

    def add_vulnerability(
        self,
        os,
        prog,
        vuln,
        ):

        self.add_application(os, prog)
        self._db[os][prog].append(vuln)
        self._numvulns += 1

    def get_num_vulnerabilities(self):
        return self._numvulns

    def get_vulnerabilities(self, os, prog):

        if not os in self._db or not prog in self._db[os]:
            return []
        return self._db[os][prog]

    def get_operating_systems(self):
        return self._db.keys()

    def get_applications(self, os):
        return self._db[os].keys()

    def cleanup(self):
        oss = self._db.keys()
        for os in oss:
            apps = self._db[os].keys()
            for app in apps:
                if len(self._db[os][app]) == 0:
                    del self._db[os][app]
            if len(self._db[os]) == 0:
                del self._db[os]

    def remove_vulnerability(
        self,
        os,
        prog,
        vuln,
        ):

        newvuln = []
        for vul in self._db[os][prog]:
            if vul != vuln:
                newvuln.append(vul)
            else:
                self._numvulns -= 1
        self._db[os][prog] = newvuln
        self.cleanup()


class Configuration(object):

    def __init__(self, name, os):

        self._os = os
        self._name = name
        self._vuln = {}

    def add_application(self, prog):
        if not prog in self._vuln:
            self._vuln[prog] = []

    def add_vulnerability(self, prog, vuln):
        self.add_application(prog)
        self._vuln[prog].append(vuln)

    def get_applications(self):
        return self._vuln.keys()

    def get_vulnerabilities(self):
        res = []
        for app in self._vuln:
            for vul in self._vuln[app]:
                res.append((self._os, app, vul))
        return res


class Template(object):

    def __init__(self, name=''):
        self._name = name
        self._confs_by_os = {}

    def add_os(self, conf, os):
        if not conf in self._confs_by_os:
            self._confs_by_os[conf] = Configuration('%s-%s'
                    % (self._name, conf), os)

    def add_application(
        self,
        conf,
        os,
        app,
        ):

        self.add_os(conf, os)
        self._confs_by_os[conf].add_application(app)

    def add_vulnerability(
        self,
        conf,
        os,
        app,
        vuln,
        ):

        self.add_application(conf, os, app)
        self._confs_by_os[conf].add_vulnerability(app, vuln)

    def get_configurations(self):
        return self._confs_by_os.values()


class Host(object):

    def __init__(self, name, template):
        self._name = name
        self._tmpl = template

    def get_name(self):
        return self._name

    def get_possible_configurations(self):
        return self._tmpl.get_configurations()


class NetEntity(object):

    def __init__(self, id):
        self.id = id

    def is_connected(
        self,
        i,
        j,
        port,
        ):

        return False

    def contains(self, i):
        return False

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    def short_repr(self):
        return ''


class SubNetwork(NetEntity):

    def __init__(self, id):
        NetEntity.__init__(self, id)
        self._hosts = []

    def add_host(self, host):
        self._hosts.append(host)

    def get_host(self, i):
        return self._hosts[i]

    def size(self):
        return len(self._hosts)

    def get_all_hosts(self):
        return self._hosts

    def is_connected(
        self,
        i,
        j,
        port,
        ):

        return i in self._hosts and j in self._hosts

    def contains(self, i):
        return i in self._hosts

    def __repr__(self):
        return '%r' % self._hosts

    def short_repr(self):
        return '%r' % self._hosts

    def in_root(self, i):
        return i in self._hosts

    def dump_graphviz_nets(self):
        print ' ' * 4 + 'net%d [label="%r"]' % (self.id, self._hosts)

    def dump_graphviz_edges(self):
        pass

    def get_root(self):
        return self


class Network(NetEntity):

    def __init__(self, id):
        NetEntity.__init__(self, id)
        self._id_to_net = {}
        self._nets = {}
        self._tip = []
        self._graph = {}
        self._numnets = 0
        self._total_hosts = 0

    def get_root(self):
        return self._id_to_net[0]

    def in_root(self, i):
        return self._id_to_net[0].in_root(i)

    def add_net(self, net):
        if net in self._nets:
            return self._nets[net]
        self._nets[net] = self._numnets
        self._tip.append(self._numnets)
        self._id_to_net[self._numnets] = net
        self._numnets += 1
        self._total_hosts += net.size()
        return self._numnets - 1

    def add_connection(
        self,
        net1,
        net2,
        fw,
        bi=False,
        ):

        i = self.add_net(net1)
        j = self.add_net(net2)
        if not i in self._graph:
            self._graph[i] = {}
        self._graph[i][j] = fw
        if bi:
            if not j in self._graph:
                self._graph[j] = {}
            self._graph[j][i] = fw
        newtip = []
        for x in self._tip:
            if x != i:
                newtip.append(x)
        self._tip = newtip

    def get_tip_nets(self):
        return self._tip

    def get_net(self, i):
        return self._id_to_net[i]

    def size(self):
        return self._total_hosts

    def num_subnets(self):
        return len(self._nets)

    def get_subnet_to_host(self, i):
        for sub1 in self._nets:
            if sub1.contains(i):
                return sub1
        return None

    def contains(self, i):

        # return not self.get_subnet_to_host(i) is None

        for net in self._nets:
            if net.contains(i):
                return True
        return False

    def dump_graphviz_nets(self):
        for net in self._nets:
            net.dump_graphviz_nets()

    def dump_graphviz_edges(self):
        for net in self._nets:
            net.dump_graphviz_edges()
        for i in self._graph:
            for j in self._graph[i]:
                net1 = self._id_to_net[i].get_root()
                net2 = self._id_to_net[j].get_root()
                print ' ' * 4 + 'net%d -> net%d [label="%r"]' \
                    % (net1.id, net2.id, self._graph[i][j])

    def is_connected(
        self,
        i,
        j,
        port,
        ):

        sub1 = self.get_subnet_to_host(i)
        sub2 = self.get_subnet_to_host(j)
        if sub1 is None or sub2 is None:
            return False
        if sub1 == sub2:
            return sub1.is_connected(i, j, port)
        if not sub1.in_root(i) or not sub2.in_root(j):
            return False
        I = self._nets[sub1]
        J = self._nets[sub2]
        if not I in self._graph or not J in self._graph[I]:
            return False
        return self._graph[I][J].allows(port)

    def short_repr(self):
        if len(self._nets) == 0:
            return '[]'
        return self._id_to_net[0].short_repr()

    def __repr__(self):
        res = ''
        for i in self._graph:
            for j in self._graph[i]:
                res += '%s ---{ %r }---> %s\n' \
                    % (self._id_to_net[i].short_repr(),
                       self._graph[i][j],
                       self._id_to_net[j].short_repr())
        for i in self._tip:
            res += '%r ---|\n' % self._id_to_net[i]
        if res.endswith('\n'):
            res = res[:-1]
        return res


class Firewall(object):

    def allows(self, port):
        return False

    def __repr__(self):
        return '-'


class EmptyFirewall(Firewall):

    def allows(self, port):
        return True

    def __repr__(self):
        return '*'


class SelectiveFirewall(Firewall):

    def __init__(self, ports):
        self._ports = ports

    def allows(self, port):
        return port in self._ports

    def __repr__(self):
        return '%r' % self._ports


class CoreSecGenerator(object):

    '''
    hosts -- number of hosts
    templates -- list of templates, each host is assigned exactly one template
    exposed -- number of hosts that go into the exposed network
    sensitive -- number of hosts that go into the sensitive network
    user -- number of hosts that go into the user network
    seed -- seed for rng
    '''

    def __init__(
        self,
        edbs,
        hosts,
        exploits,
        fw_inc_exposed,
        fw_inc_sensitive,
        fw_inc_user,
        templates,
        subnet_size=5,
        exposed=1,
        sensitive=1,
        per=20,
        seed=0,
        comment=None,
        ):

        self._comment = comment

        random.seed(seed)
        s_template = random.randint(0, sys.maxint)
        s_network = random.randint(0, sys.maxint)

        random.seed(s_template)
        self._templates = []
        i = 0
        for edb in edbs:
            self._templates.append(self.generate_templates(edb,
                                   exploits[i], templates[i]))
            i += 1
        self._hosts = []
        x1 = 0
        x2 = 0
        x3 = 0
        for i in range(hosts):
            if i % per < exposed:
                j = x1 % templates[0]  # uniformly distribute templates
                self._hosts.append(Host('host%d' % i,
                                   self._templates[0][j]))
                x1 += 1
            elif i % per < exposed + sensitive:
                j = x2 % templates[1]  # uniformly distribute templates
                self._hosts.append(Host('host%d' % i,
                                   self._templates[1][j]))
                x2 += 1
            else:
                j = x3 % templates[2]  # uniformly distribute templates
                self._hosts.append(Host('host%d' % i,
                                   self._templates[2][j]))
                x3 += 1
        self._hosts.append(Host('internet', Template()))

        random.seed(s_network)
        self._net = self.generate_network(
            hosts,
            fw_inc_exposed,
            fw_inc_sensitive,
            fw_inc_user,
            subnet_size,
            exposed,
            sensitive,
            per,
            )

        self._net.get_net(0).add_host(hosts)

        # self._goal = self.select_goal_hosts(sensitive_goal, user_goal)

        self._goal = []
        sensitive_net = self._net.get_net(2)
        user_net = self._net.get_net(3)
        if sensitive_net.size() > 0:
            self._goal.append(sensitive_net.get_host(0))
        if user_net.size() > 0:
            tip_net = user_net.get_net(user_net.get_tip_nets()[-1])
            self._goal.append(tip_net.get_all_hosts()[0])

        # for i in range(len(user_net.get_tip_nets())):
        #    self._goal.append(user_net.get_net(user_net.get_tip_nets()[i]).get_all_hosts()[0])

        # print "internet"
        # print self._net.get_net(0)
        # print
        # print "exposed"
        # print self._net.get_net(1)
        # print
        # print "sensitive"
        # print self._net.get_net(2)
        # print
        # print "user"
        # print user_net
        # print
        # print "ALL"
        # print self._net

        #print 'digraph {'
        #self._net.dump_graphviz_nets()
        #self._net.dump_graphviz_edges()
        #print '}'

    def generate_templates(
        self,
        dbs,
        exploits,
        templates,
        ):

        num_exploits = [0 for i in range(templates)]
        for i in range(exploits):
            num_exploits[i % templates] += 1  # uniformly distribute exploits over templates
        result = []
        seeds = [random.randint(0, sys.maxint) for i in
                 range(templates)]
        for i in range(templates):
            random.seed(seeds[i])
            db = copy.deepcopy(dbs.get(i))
            templ = Template('tmpl%d' % i)
            for os in db.get_operating_systems():
                for app in db.get_applications(os):
                    templ.add_application(os.get_name(), os, app)
            if db.get_num_vulnerabilities() < num_exploits[i]:
                sys.stderr.write('WARNING: The selected exploit database does not contain enough exploits!\n'
                                 )
                num_exploits[i] = db.get_num_vulnerabilities()
            db.cleanup()  # in the following ignore configurations without vulnerabilities
            while num_exploits[i] > 0:
                os = random.randint(0, len(db.get_operating_systems())
                                    - 1)
                os = db.get_operating_systems()[os]
                app = random.randint(0, len(db.get_applications(os))
                        - 1)
                app = db.get_applications(os)[app]
                vul = random.randint(0, len(db.get_vulnerabilities(os,
                        app)) - 1)
                vul = db.get_vulnerabilities(os, app)[vul]
                db.remove_vulnerability(os, app, vul)
                templ.add_vulnerability(os.get_name(), os, app, vul)
                num_exploits[i] -= 1
            result.append(templ)
        return result

    def generate_network(
        self,
        hosts,
        fw_inc_exposed,
        fw_inc_sensitive,
        fw_inc_user,
        subnet_size,
        exposed,
        sensitive,
        per,
        ):

        netid = 5
        internet = SubNetwork(0)
        net_exposed = SubNetwork(1)
        net_sensitive = SubNetwork(2)
        subnet_user = SubNetwork(3)
        net_user = Network(4)
        N = subnet_size
        B = None
        C = 0

        for i in range(hosts):
            if i % per < exposed:
                net_exposed.add_host(i)
            elif i % per < exposed + sensitive:
                net_sensitive.add_host(i)
            else:
                subnet_user.add_host(i)
                if subnet_user.size() >= N:
                    n = net_user.num_subnets()
                    if n == 0:
                        net_user.add_net(subnet_user)
                    else:
                        j = random.randint(0, n - 1)
                        net_user.add_connection(net_user.get_net(j),
                                subnet_user, EmptyFirewall(),
                                bi=True)
                    subnet_user = SubNetwork(netid)
                    N = subnet_size
                    netid += 1
        if subnet_user.size() > 0:
            n = net_user.num_subnets()
            if n == 0:
                net_user.add_net(subnet_user)
            else:
                j = random.randint(0, n - 1)
                net_user.add_connection(net_user.get_net(j),
                        subnet_user, EmptyFirewall(), bi=True)
            netid += 1

        net = Network(netid)
        net.add_net(internet)
        net.add_net(net_exposed)
        net.add_net(net_sensitive)
        net.add_net(net_user)
        net.add_connection(internet, net_exposed, fw_inc_exposed)
        net.add_connection(net_exposed, net_sensitive, fw_inc_sensitive)
        net.add_connection(net_exposed, net_user, fw_inc_user)
        net.add_connection(net_user, net_sensitive, fw_inc_sensitive)

        # net.add_connection(net_user, net_exposed, EmptyFirewall())

        return net

    def _exploit_to_string(
        self,
        h,
        os,
        app,
        vul,
        ):

        return '%s_%s_%s_%s' % (h.get_name(), os.get_name(),
                                app.get_name(), vul.get_name())

    def generate_pddl_domain(self, fdom, dom_name='CoreSec'):
        with open(fdom, 'w') as f:
            if not self._comment is None:
                f.write(';;; %s\n' % self._comment)
            f.write('(define (domain %s)\n' % dom_name)
            f.write('''(:requirements :strips :probabilistic-effects)

''')
            f.write('''(:types host port)

''')
            f.write('(:predicates\n')
            f.write(' ' * 4 + '(hacl ?src ?dest - host ?p - port)\n')
            f.write(' ' * 4 + '(controlling ?h - host)\n')
            ports = []
            exploits = []
            for h in self._hosts:
                for conf in h.get_possible_configurations():
                    for (os, app, vul) in conf.get_vulnerabilities():
                        exploits.append(self._exploit_to_string(h, os,
                                app, vul))
                        ports.append(app.get_port())
            ports = set(ports)
            exploits = set(exploits)
            for exploit in exploits:
                f.write('    (expl_%s)\n' % exploit)
            f.write(''')

''')
            f.write('''(:functions (total-cost))

''')
            f.write('(:constants\n    ')
            for h in self._hosts:
                f.write('%s ' % h.get_name())
            f.write('- host\n    ')
            for p in ports:
                f.write('p%d ' % p)
            f.write('- port\n')
            f.write(''')

''')
            exploits = []
            for h in self._hosts:
                for conf in h.get_possible_configurations():
                    for (os, app, vul) in conf.get_vulnerabilities():
                        expl = self._exploit_to_string(h, os, app, vul)
                        if expl in exploits:
                            continue
                        exploits.append(expl)
                        f.write('(:action ex-%s\n'
                                % self._exploit_to_string(h, os, app,
                                vul))
                        f.write('    :parameters (?src - host)\n')
                        f.write('    :precondition (and (hacl ?src %s p%d) (controlling ?src) (expl_%s))\n'
                                 % (h.get_name(), app.get_port(), expl))
                        f.write('    :effect (and (increase (total-cost) %d) '
                                 % vul.get_cost())
                        f.write('(not (expl_%s)) ' % expl)
                        prob = PROB_DENOMINATOR
                        if SUCESS_PROBABILITIES:
                            prob = vul.get_probability()
                        if prob != PROB_DENOMINATOR:
                            f.write('(probabilistic %d/%d (and '
                                    % (prob, PROB_DENOMINATOR))
                        f.write('(controlling %s)' % h.get_name())
                        if prob != PROB_DENOMINATOR:
                            f.write('))')
                        f.write(''')
)

''')

            f.write(')\n')

    def generate_pddl_problem(
        self,
        fprob,
        dom_name='CoreSec',
        prob_name='CoreSecP',
        ):

        with open(fprob, 'w') as f:
            if not self._comment is None:
                f.write(';;; %s\n' % self._comment)
            f.write('(define (problem %s)\n' % prob_name)
            f.write('(:domain %s)\n' % dom_name)
            f.write('(:init\n')
            f.write(' ' * 4 + '(= (total-cost) 0)\n')
            f.write(' ' * 4 + '(controlling internet)\n')
            ports = []
            for i in range(len(self._hosts)):
                h = self._hosts[i]
                X = []
                for conf in h.get_possible_configurations():
                    for (_, app, _) in conf.get_vulnerabilities():
                        X.append(app.get_port())
                ports.append(set(X))
            for i in range(len(self._hosts)):
                for j in range(len(self._hosts)):
                    if i == j:
                        continue
                    for port in ports[i]:
                        if self._net.is_connected(j, i, port):
                            f.write(' ' * 4 + '(hacl %s %s p%d)\n'
                                    % (self._hosts[j].get_name(),
                                    self._hosts[i].get_name(), port))
            exploits = []
            for h in self._hosts:
                for conf in h.get_possible_configurations():
                    for (os, app, vul) in conf.get_vulnerabilities():
                        expl = self._exploit_to_string(h, os, app, vul)
                        if expl in exploits:
                            continue
                        exploits.append(expl)
                        f.write(' ' * 4 + '(expl_%s)\n' % expl)
            f.write(')\n')
            f.write('(:goal (and\n')
            for g in self._goal:
                f.write(' ' * 4 + '(controlling %s)\n'
                        % self._hosts[g].get_name())
            f.write('))\n')
            f.write('(:metric minimize (total-cost))\n')
            f.write(')')


def generate_db(
    os_apps,
    exploits,
    mp,
    xp,
    mc=1,
    xc=1,
    ):
    db = ExploitDatabase()
    for os in os_apps:
        for app in os_apps[os]:
            db.add_application(os, app)
    os_keys = os_apps.keys()
    cexploits = 0
    while cexploits < exploits:
        os = os_keys[random.randint(0, len(os_keys) - 1)]
        app = os_apps[os][random.randint(0, len(os_apps[os]) - 1)]
        db.add_vulnerability(os, app, Vulnerability('vuln%d'
                             % cexploits, random.randint(mp, xp),
                             random.randint(mc, xc)))
        cexploits += 1
    return db


def generate_edb_core_sec_web_server(
    exploits,
    mp,
    xp,
    mc,
    xc,
    ):
    res = []
    for x in [[
        ('winNT4ser', 'IIS4', 80),
        ('winNT4ent', 'IIS4', 80),
        ('win2000adv', 'IIS5', 80),
        ('win2000pro', 'IIS5', 80),
        ('win2000ser', 'IIS5', 80),
        ('win2000adv', 'IIS5', 80),
        ], [
        ('win2000pro', 'IIS5', 80),
        ('win2000ser', 'IIS5', 80),
        ('win2000adv', 'IIS5', 80),
        ('winNT4ser', 'IIS4', 80),
        ('win2000ser', 'IIS5', 80),
        ('win2000adv', 'IIS5', 80),
        ('win2000ser', 'IIS5', 80),
        ('win2000adv', 'IIS5', 80),
        ], [
        ('win2000ser', 'IIS5', 80),
        ('win2000adv', 'IIS5', 80),
        ('winNT4ser', 'IIS4', 80),
        ('winNT4ent', 'IIS4', 80),
        ('win2000ser', 'IIS5', 80),
        ('win2000adv', 'IIS5', 80),
        ], [
        ('winNT4ser', 'IIS4', 80),
        ('winNT4ent', 'IIS4', 80),
        ('win2000ser', 'IIS5', 80),
        ('win2000adv', 'IIS5', 80),
        ('win2000ser', 'IIS5', 80),
        ('win2000adv', 'IIS5', 80),
        ]]:
        os_apps = {}
        for (os, app, port) in x:
            oss = OperatingSystem(os)
            if not oss in os_apps:
                os_apps[oss] = []
            appp = Application(app, port)
            if not appp in os_apps[oss]:
                os_apps[oss].append(appp)
            res.append(generate_db(
                os_apps,
                exploits,
                mp,
                xp,
                mc,
                xc,
                ))
    return res


def generate_edb_core_sec_sql_server(
    exploits,
    mp,
    xp,
    mc,
    xc,
    ):
    res = []
    for x in [[
        ('winNT4ser', 'SQL4', 3306),
        ('winNT4ent', 'SQL4', 3306),
        ('win2000adv', 'SQL5', 3306),
        ('win2000pro', 'SQL5', 3306),
        ('win2000ser', 'SQL5', 3306),
        ('win2000adv', 'SQL5', 3306),
        ], [
        ('win2000pro', 'SQL5', 3306),
        ('win2000ser', 'SQL5', 3306),
        ('win2000adv', 'SQL5', 3306),
        ('winNT4ser', 'SQL4', 3306),
        ('win2000ser', 'SQL5', 3306),
        ('win2000adv', 'SQL5', 3306),
        ('win2000ser', 'SQL5', 3306),
        ('win2000adv', 'SQL5', 3306),
        ], [
        ('win2000ser', 'SQL5', 3306),
        ('win2000adv', 'SQL5', 3306),
        ('winNT4ser', 'SQL4', 3306),
        ('winNT4ent', 'SQL4', 3306),
        ('win2000ser', 'SQL5', 3306),
        ('win2000adv', 'SQL5', 3306),
        ], [
        ('winNT4ser', 'SQL4', 3306),
        ('winNT4ent', 'SQL4', 3306),
        ('win2000ser', 'SQL5', 3306),
        ('win2000adv', 'SQL5', 3306),
        ('win2000ser', 'SQL5', 3306),
        ('win2000adv', 'SQL5', 3306),
        ]]:

        os_apps = {}
        for (os, app, port) in x:
            oss = OperatingSystem(os)
            if not oss in os_apps:
                os_apps[oss] = []
            appp = Application(app, port)
            if not appp in os_apps[oss]:
                os_apps[oss].append(appp)
            res.append(generate_db(
                os_apps,
                exploits,
                mp,
                xp,
                mc,
                xc,
                ))
    return res


def generate_edb_core_sec_user(
    exploits,
    mp,
    xp,
    mc,
    xc,
    ):
    res = []
    for x in [
        [('winNT4ser', 'MSRPC4', 445)],
        [('win2000pro', 'MSRPC5', 445)],
        [('win2000ser', 'MSRPC5', 445)],
        [('win2000ser', 'MSRPC5', 445)],
        [('win2000ser', 'MSRPC5', 445)],
        [('win2000adv', 'MSRPC5', 445)],
        [('win2000ser', 'MSRPC5', 445)],
        ]:

        os_apps = {}
        for (os, app, port) in x:
            oss = OperatingSystem(os)
            if not oss in os_apps:
                os_apps[oss] = []
            appp = Application(app, port)
            if not appp in os_apps[oss]:
                os_apps[oss].append(appp)
            res.append(generate_db(
                os_apps,
                exploits,
                mp,
                xp,
                mc,
                xc,
                ))
    return res


def generate_coresec_dbs(
    exploits,
    seed=0,
    mp=1,
    xp=PROB_DENOMINATOR,
    mc=1,
    xc=1,
    ):

    class CoreSecEDB(object):

        def __init__(
            self,
            web,
            sql,
            user,
            ):

            self.web = web
            self.sql = sql
            self.user = user
            self.empty = ExploitDatabase()

        def get(self, i):
            if i % 3 == 0:
                return self.web[i % 4]
            elif i % 3 == 1:
                return self.sql[i % 4]
            else:
                return self.user[i % 7]

    random.seed(seed)
    return CoreSecEDB(generate_edb_core_sec_web_server(exploits, mp,
                      xp, mc, xc),
                      generate_edb_core_sec_sql_server(exploits, mp,
                      xp, mc, xc), generate_edb_core_sec_user(exploits,
                      mp, xp, mc, xc))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('hosts', type=int, help='number of hosts')
    parser.add_argument('exploits_exposed', type=int, help='number of exploits in exposed network')
    parser.add_argument('exploits_sensitive', type=int, help='number of exploits in sensitive network')
    parser.add_argument('exploits_user', type=int, help='number of exploits in user network')
    parser.add_argument('domain', type=str,
                        help='path to resulting domain file')
    parser.add_argument('problem', type=str,
                        help='path to resulting problem file')
    parser.add_argument(
        '--texposed',
        type=int,
        action='store',
        dest='templates_exposed',
        default=1,
        help='number of host templates to be generated in exposed network',
        )
    parser.add_argument(
        '--tsensitive',
        type=int,
        action='store',
        dest='templates_sensitive',
        default=1,
        help='number of host templates to be generated in sensitive network',
        )
    parser.add_argument(
        '--tuser',
        type=int,
        action='store',
        dest='templates_user',
        default=11,
        help='number of host templates to be generated in user network',
        )
    parser.add_argument(
        '--distribution',
        type=int,
        action='store',
        dest='distribution',
        default=40,
        help='distribute per N hosts one host to sensitive, one host to exposed, and N - 2 to user'
            ,
        )
    parser.add_argument(
        '--subnet',
        type=int,
        action='store',
        dest='subnet',
        default=5,
        help='subnet size of user network',
        )
    parser.add_argument(
        '--seed',
        type=int,
        action='store',
        dest='seed',
        default=0,
        help='seed for RNG',
        )
    parser.add_argument(
        '--genexplts',
        type=int,
        action='store',
        dest='genexplts',
        default=100,
        help='number of exploits to be generated',
        )
    args = parser.parse_args()

    explts = args.genexplts
    if explts < args.exploits_exposed + args.exploits_sensitive + args.exploits_user:
        explts = args.exploits

    t1 = args.templates_exposed
    t2 = args.templates_sensitive
    t3 = args.templates_user
    e1 = args.exploits_exposed
    e2 = args.exploits_sensitive
    e3 = args.exploits_user

    random.seed(args.seed)
    s_db = random.randint(0, sys.maxint)
    s_gen = random.randint(0, sys.maxint)
    inst = CoreSecGenerator(  # 10, 15),
                              # 5, 10),
                              # 1, 5),
        [generate_coresec_dbs(
            explts,
            s_db,
            15,
            40,
            1,
            1,
            ), generate_coresec_dbs(
            explts,
            s_db,
            35,
            60,
            1,
            1,
            ), generate_coresec_dbs(
            explts,
            s_db,
            50,
            80,
            1,
            1,
            )],
        args.hosts,
        [e1, e2, e3],
        SelectiveFirewall([80, 3306, 445]),
        SelectiveFirewall([80, 3306, 445]),
        SelectiveFirewall([80, 3306, 445]),
        [t1, t2, t3],
        subnet_size=args.subnet,
        per=args.distribution,
        seed=s_gen,
        comment='H=%d E=%r T=%r Sub=%d 1-per=%d Seed=%d(%d)' % (
            args.hosts,
            [e1, e2, e3],
            [t1, t2, t3],
            args.subnet,
            args.distribution,
            args.seed,
            s_gen,
            ),
        )
    inst.generate_pddl_domain(args.domain)
    inst.generate_pddl_problem(args.problem)

